// Copyright (c) 2021 - 2022 Buijs Software
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

import "dart:core";
import "dart:io";

import "package:squint_json/src/cli/const.dart";
import "package:squint_json/src/cli/generate_arguments.dart";
import "package:squint_json/src/cli/generate_serializer.dart";
import "package:test/test.dart";

void main() {
  test("When the input File does not exist then a Result NOK is returned", () {
    // given:
    final arguments = {
      GenerateArgs.type: generateArgumentTypeValueSerializer,
      GenerateArgs.input: "doesNotExist.dart"
    };

    // when:
    final result = arguments.serializers;
    expect(result.isOk, false);
    expect(result.nok != null, true);
    expect(result.nok!.length, 1);
    expect(result.nok![0].startsWith("File does not exist:"), true);
    expect(result.nok![0].endsWith("doesNotExist.dart"), true);
  });

  test(
      "When the input File exists but is not a Dart class then a Result NOK is returned",
      () {
    // setup:
    final notDartFile =
        File("${Directory.systemTemp.path}${Platform.pathSeparator}dotDart.txt")
          ..createSync(recursive: true);

    // given:
    final arguments = {
      GenerateArgs.type: generateArgumentTypeValueSerializer,
      GenerateArgs.input: notDartFile.path,
    };

    // when:
    final result = arguments.serializers;
    expect(result.isOk, false);
    expect(result.nok != null, true);
    expect(result.nok!.length, 1);
    expect(result.nok![0].startsWith("File is not a .dart File:"), true);
    expect(result.nok![0].endsWith("dotDart.txt"), true);
  });

  test(
      "When the input File Dart content is invalid then a Result NOK is returned",
      () {
    // setup:
    final invalidDart = File(
        "${Directory.systemTemp.path}${Platform.pathSeparator}invalid.dart")
      ..createSync(recursive: true)
      ..writeAsStringSync("""yuck""");

    // given:
    final arguments = {
      GenerateArgs.type: generateArgumentTypeValueSerializer,
      GenerateArgs.input: invalidDart.path,
    };

    // when:
    final result = arguments.serializers;
    expect(result.isOk, false);
    expect(result.nok != null, true);
    expect(result.nok!.length, 1);
    expect(result.nok![0].startsWith("Failed to analyze .dart File:"), true);
    expect(result.nok![0].endsWith("invalid.dart"), true);
  });

  test("When the output File already exists then a Result NOK is returned", () {
    // setup:
    final exampleDart = File("example_dataclass.dart")
      ..createSync(recursive: true)
      ..writeAsStringSync("""
        import 'package:squint_json/squint_json.dart';

        /// Autogenerated data class by Squint.
        @squint
        class Example {
          const Example({
            required this.id,
          });
        
          @JsonValue("id")
          final int id;
          
        }
      """);

    // setup: output file exists
    final extensionsFile = File("example_extensions.dart")
      ..createSync(recursive: true);

    // given:
    final arguments = {
      GenerateArgs.type: generateArgumentTypeValueSerializer,
      GenerateArgs.input: exampleDart.path,
    };

    // when:
    final result = arguments.serializers;

    // cleanup:
    exampleDart.delete();
    extensionsFile.delete();

    // then:
    expect(result.isOk, false);
    expect(result.nok != null, true);
    expect(result.nok!.length, 2);
    expect(
        result.nok![0].startsWith(
            "Failed to write generated code because File already exists"),
        true);
    expect(result.nok![0].endsWith("example_extensions.dart."), true);
    expect(result.nok![1],
        "Use '--$generateArgumentOverwrite true' to allow overwriting existing files.");
  });
}
